@@ -40,17 +40,10 @@
 #include "gdbusutils.h"
 #include "gmountoperationdbus.h"
 #include <gio/gio.h>
+#include "metatree.h"
 
 static void g_daemon_file_file_iface_init (GFileIface       *iface);
 
-struct _GDaemonFile
-{
-  GObject parent_instance;
-
-  GMountSpec *mount_spec;
-  char *path;
-};
-
 static void g_daemon_file_read_async (GFile *file,
 				      int io_priority,
 				      GCancellable *cancellable,
@@ -88,6 +81,29 @@ g_daemon_file_init (GDaemonFile *daemon_file)
 {
 }
 
+static guint32
+get_pid_for_file (GFile *file)
+{
+  guint32 pid;
+
+  pid = 0;
+  if (file == NULL)
+    goto out;
+
+  /* The fuse client sets this to convey the pid of the client - see
+   * set_pid_for_file() in gvfsfusedaemon.c
+   */
+  pid = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (file), "gvfs-fuse-client-pid"));
+  if (pid != 0)
+    goto out;
+
+  /* otherwise assume the client is this process */
+  pid = (guint32) getpid ();
+
+ out:
+  return pid;
+}
+
 GFile *
 g_daemon_file_new (GMountSpec *mount_spec,
 		   const char *path)
@@ -580,7 +596,7 @@ async_path_call_done (DBusMessage *reply,
   if (io_error != NULL)
     {
       g_simple_async_result_set_from_error (data->result, io_error);
-      g_simple_async_result_complete (data->result);
+      _g_simple_async_result_complete_with_cancellable (data->result, data->cancellable);
       async_path_call_free (data);
     }
   else
@@ -613,7 +629,7 @@ do_async_path_call_callback (GMountInfo *mount_info,
   if (error != NULL)
     {
       g_simple_async_result_set_from_error (data->result, error);      
-      g_simple_async_result_complete (data->result);
+      _g_simple_async_result_complete_with_cancellable (data->result, data->cancellable);
       async_path_call_free (data);
       return;
     }
@@ -709,7 +725,7 @@ g_daemon_file_enumerate_children (GFile      *file,
   DBusConnection *connection;
   char *uri;
 
-  enumerator = g_daemon_file_enumerator_new (file);
+  enumerator = g_daemon_file_enumerator_new (file, attributes);
   obj_path = g_daemon_file_enumerator_get_object_path (enumerator);
 
 
@@ -746,6 +762,70 @@ g_daemon_file_enumerate_children (GFile      *file,
   return NULL;
 }
 
+
+static gboolean
+enumerate_keys_callback (const char *key,
+			 MetaKeyType type,
+			 gpointer value,
+			 gpointer user_data)
+{
+  GFileInfo  *info = user_data;
+  char *attr;
+
+  attr = g_strconcat ("metadata::", key, NULL);
+
+  if (type == META_KEY_TYPE_STRING)
+    g_file_info_set_attribute_string (info, attr, (char *)value);
+  else
+    g_file_info_set_attribute_stringv (info, attr, (char **)value);
+
+  g_free (attr);
+
+  return TRUE;
+}
+
+static void
+add_metadata (GFile *file,
+	      const char *attributes,
+	      GFileInfo *info)
+{
+  GDaemonFile *daemon_file;
+  GFileAttributeMatcher *matcher;
+  const char *first;
+  char *treename;
+  gboolean all;
+  MetaTree *tree;
+
+  daemon_file = G_DAEMON_FILE (file);
+
+  matcher = g_file_attribute_matcher_new (attributes);
+  all = g_file_attribute_matcher_enumerate_namespace (matcher, "metadata");
+
+  first = NULL;
+  if (!all)
+    {
+      first = g_file_attribute_matcher_enumerate_next (matcher);
+
+      if (first == NULL)
+	{
+	  g_file_attribute_matcher_unref (matcher);
+	  return; /* No match */
+	}
+    }
+
+  treename = g_mount_spec_to_string (daemon_file->mount_spec);
+  tree = meta_tree_lookup_by_name (treename, FALSE);
+  g_free (treename);
+
+  g_file_info_set_attribute_mask (info, matcher);
+  meta_tree_enumerate_keys (tree, daemon_file->path,
+			    enumerate_keys_callback, info);
+  g_file_info_unset_attribute_mask (info);
+
+  meta_tree_unref (tree);
+  g_file_attribute_matcher_unref (matcher);
+}
+
 static GFileInfo *
 g_daemon_file_query_info (GFile                *file,
 			  const char           *attributes,
@@ -790,6 +870,9 @@ g_daemon_file_query_info (GFile                *file,
 
   info = _g_dbus_get_file_info (&iter, error);
 
+  if (info)
+    add_metadata (file, attributes, info);
+
  out:
   dbus_message_unref (reply);
   return info;
@@ -802,9 +885,11 @@ query_info_async_cb (DBusMessage *reply,
 		     GCancellable *cancellable,
 		     gpointer callback_data)
 {
+  const char *attributes = callback_data;
   DBusMessageIter iter;
   GFileInfo *info;
   GError *error;
+  GFile *file;
 
   info = NULL;
   
@@ -814,7 +899,7 @@ query_info_async_cb (DBusMessage *reply,
       g_simple_async_result_set_error (result,
 				       G_IO_ERROR, G_IO_ERROR_FAILED,
 				       _("Invalid return value from query_info"));
-      g_simple_async_result_complete (result);
+      _g_simple_async_result_complete_with_cancellable (result, cancellable);
       return;
     }
 
@@ -824,12 +909,16 @@ query_info_async_cb (DBusMessage *reply,
     {
       g_simple_async_result_set_from_error (result, error);
       g_error_free (error);
-      g_simple_async_result_complete (result);
+      _g_simple_async_result_complete_with_cancellable (result, cancellable);
       return;
     }
 
+  file = G_FILE (g_async_result_get_source_object (G_ASYNC_RESULT (result)));
+  add_metadata (file, attributes, info);
+  g_object_unref (file);
+
   g_simple_async_result_set_op_res_gpointer (result, info, g_object_unref);
-  g_simple_async_result_complete (result);
+  _g_simple_async_result_complete_with_cancellable (result, cancellable);
 }
 
 static void
@@ -851,7 +940,7 @@ g_daemon_file_query_info_async (GFile                      *file,
 		      G_VFS_DBUS_MOUNT_OP_QUERY_INFO,
 		      cancellable,
 		      callback, user_data,
-		      query_info_async_cb, NULL, NULL,
+		      query_info_async_cb, g_strdup (attributes), g_free,
 		      DBUS_TYPE_STRING, &attributes,
 		      DBUS_TYPE_UINT32, &dbus_flags,
 		      DBUS_TYPE_STRING, &uri,
@@ -877,6 +966,7 @@ g_daemon_file_query_info_finish (GFile                      *file,
 
 typedef struct {
   GSimpleAsyncResult *result;
+  GCancellable *cancellable;
   gboolean can_seek;
 } GetFDData;
 
@@ -899,7 +989,7 @@ read_async_get_fd_cb (int fd,
       g_simple_async_result_set_op_res_gpointer (data->result, stream, g_object_unref);
     }
 
-  g_simple_async_result_complete (data->result);
+  _g_simple_async_result_complete_with_cancellable (data->result, data->cancellable);
 
   g_object_unref (data->result);
   g_free (data);
@@ -924,7 +1014,7 @@ read_async_cb (DBusMessage *reply,
       g_simple_async_result_set_error (result,
 				       G_IO_ERROR, G_IO_ERROR_FAILED,
 				       _("Invalid return value from open"));
-      g_simple_async_result_complete (result);
+      _g_simple_async_result_complete_with_cancellable (result, cancellable);
       return;
     }
   
@@ -943,11 +1033,16 @@ g_daemon_file_read_async (GFile *file,
 			  GAsyncReadyCallback callback,
 			  gpointer callback_data)
 {
+  guint32 pid;
+
+  pid = get_pid_for_file (file);
+
   do_async_path_call (file,
 		      G_VFS_DBUS_MOUNT_OP_OPEN_FOR_READ,
 		      cancellable,
 		      callback, callback_data,
 		      read_async_cb, NULL, NULL,
+                      DBUS_TYPE_UINT32, &pid,
 		      0);
 }
 
@@ -977,11 +1072,15 @@ g_daemon_file_read (GFile *file,
   DBusMessage *reply;
   guint32 fd_id;
   dbus_bool_t can_seek;
+  guint32 pid;
+
+  pid = get_pid_for_file (file);
 
   reply = do_sync_path_call (file, 
 			     G_VFS_DBUS_MOUNT_OP_OPEN_FOR_READ,
 			     NULL, &connection,
 			     cancellable, error,
+                             DBUS_TYPE_UINT32, &pid,
 			     0);
   if (reply == NULL)
     return NULL;
@@ -1026,6 +1125,9 @@ g_daemon_file_append_to (GFile *file,
   dbus_bool_t make_backup;
   guint32 dbus_flags;
   char *etag;
+  guint32 pid;
+
+  pid = get_pid_for_file (file);
 
   mode = 1;
   etag = "";
@@ -1040,6 +1142,7 @@ g_daemon_file_append_to (GFile *file,
 			     DBUS_TYPE_STRING, &etag,
 			     DBUS_TYPE_BOOLEAN, &make_backup,
 			     DBUS_TYPE_UINT32, &dbus_flags,
+                             DBUS_TYPE_UINT32, &pid,
 			     0);
   if (reply == NULL)
     return NULL;
@@ -1085,6 +1188,9 @@ g_daemon_file_create (GFile *file,
   dbus_bool_t make_backup;
   char *etag;
   guint32 dbus_flags;
+  guint32 pid;
+
+  pid = get_pid_for_file (file);
 
   mode = 0;
   etag = "";
@@ -1099,6 +1205,7 @@ g_daemon_file_create (GFile *file,
 			     DBUS_TYPE_STRING, &etag,
 			     DBUS_TYPE_BOOLEAN, &make_backup,
 			     DBUS_TYPE_UINT32, &dbus_flags,
+                             DBUS_TYPE_UINT32, &pid,
 			     0);
   if (reply == NULL)
     return NULL;
@@ -1145,6 +1252,9 @@ g_daemon_file_replace (GFile *file,
   guint64 initial_offset;
   dbus_bool_t dbus_make_backup;
   guint32 dbus_flags;
+  guint32 pid;
+
+  pid = get_pid_for_file (file);
 
   mode = 2;
   dbus_make_backup = make_backup;
@@ -1161,6 +1271,7 @@ g_daemon_file_replace (GFile *file,
 			     DBUS_TYPE_STRING, &etag,
 			     DBUS_TYPE_BOOLEAN, &dbus_make_backup,
 			     DBUS_TYPE_UINT32, &dbus_flags,
+                             DBUS_TYPE_UINT32, &pid,
 			     0);
   if (reply == NULL)
     return NULL;
@@ -1236,7 +1347,7 @@ mount_mountable_async_cb (DBusMessage *reply,
       g_simple_async_result_set_error (result,
 				       G_IO_ERROR, G_IO_ERROR_FAILED,
 				       _("Invalid return value from call"));
-      g_simple_async_result_complete (result);
+      _g_simple_async_result_complete_with_cancellable (result, cancellable);
 
       return;
     }
@@ -1253,7 +1364,7 @@ mount_mountable_async_cb (DBusMessage *reply,
 	  g_simple_async_result_set_error (result,
 					   G_IO_ERROR, G_IO_ERROR_FAILED,
 					   _("Invalid return value from call"));
-	  g_simple_async_result_complete (result);
+          _g_simple_async_result_complete_with_cancellable (result, cancellable);
 	  return;
 	}
       
@@ -1275,7 +1386,7 @@ mount_mountable_async_cb (DBusMessage *reply,
       
     }
   else
-    g_simple_async_result_complete (result);
+    _g_simple_async_result_complete_with_cancellable (result, cancellable);
 }
 
 static void
@@ -1325,6 +1436,109 @@ g_daemon_file_mount_mountable_finish (GFile               *file,
   return NULL;
 }
 
+static void
+start_mountable_async_cb (DBusMessage *reply,
+			  DBusConnection *connection,
+			  GSimpleAsyncResult *result,
+			  GCancellable *cancellable,
+			  gpointer callback_data)
+{
+  _g_simple_async_result_complete_with_cancellable (result, cancellable);
+}
+
+static void
+g_daemon_file_start_mountable (GFile               *file,
+			       GDriveStartFlags     flags,
+			       GMountOperation     *mount_operation,
+			       GCancellable        *cancellable,
+			       GAsyncReadyCallback  callback,
+			       gpointer             user_data)
+{
+  GMountSource *mount_source;
+  const char *dbus_id, *obj_path;
+
+  mount_source = g_mount_operation_dbus_wrap (mount_operation, _g_daemon_vfs_get_async_bus ());
+
+  dbus_id = g_mount_source_get_dbus_id (mount_source);
+  obj_path = g_mount_source_get_obj_path (mount_source);
+
+  if (mount_operation)
+    g_object_ref (mount_operation);
+
+  do_async_path_call (file,
+		      G_VFS_DBUS_MOUNT_OP_START_MOUNTABLE,
+		      cancellable,
+		      callback, user_data,
+		      start_mountable_async_cb,
+		      mount_operation, mount_operation ? g_object_unref : NULL,
+		      DBUS_TYPE_STRING, &dbus_id,
+		      DBUS_TYPE_OBJECT_PATH, &obj_path,
+		      0);
+
+  g_object_unref (mount_source);
+}
+
+static gboolean
+g_daemon_file_start_mountable_finish (GFile               *file,
+				      GAsyncResult        *result,
+				      GError             **error)
+{
+  return TRUE;
+}
+
+static void
+stop_mountable_async_cb (DBusMessage *reply,
+                         DBusConnection *connection,
+                         GSimpleAsyncResult *result,
+                         GCancellable *cancellable,
+                         gpointer callback_data)
+{
+  _g_simple_async_result_complete_with_cancellable (result, cancellable);
+}
+
+static void
+g_daemon_file_stop_mountable (GFile               *file,
+                              GMountUnmountFlags   flags,
+                              GMountOperation     *mount_operation,
+                              GCancellable        *cancellable,
+                              GAsyncReadyCallback  callback,
+                              gpointer             user_data)
+{
+  guint32 dbus_flags;
+  GMountSource *mount_source;
+  const char *dbus_id, *obj_path;
+
+  mount_source = g_mount_operation_dbus_wrap (mount_operation, _g_daemon_vfs_get_async_bus ());
+
+  dbus_id = g_mount_source_get_dbus_id (mount_source);
+  obj_path = g_mount_source_get_obj_path (mount_source);
+
+  if (mount_operation)
+    g_object_ref (mount_operation);
+
+  dbus_flags = flags;
+  do_async_path_call (file,
+		      G_VFS_DBUS_MOUNT_OP_STOP_MOUNTABLE,
+		      cancellable,
+		      callback, user_data,
+		      stop_mountable_async_cb,
+		      mount_operation, mount_operation ? g_object_unref : NULL,
+		      DBUS_TYPE_UINT32, &dbus_flags,
+                      DBUS_TYPE_STRING, &dbus_id, DBUS_TYPE_OBJECT_PATH, &obj_path,
+		      0);
+
+  g_object_unref (mount_source);
+}
+
+static gboolean
+g_daemon_file_stop_mountable_finish (GFile               *file,
+                                     GAsyncResult        *result,
+                                     GError             **error)
+{
+  return TRUE;
+}
+
+
 static void
 eject_mountable_async_cb (DBusMessage *reply,
 			  DBusConnection *connection,
@@ -1332,7 +1546,49 @@ eject_mountable_async_cb (DBusMessage *reply,
 			  GCancellable *cancellable,
 			  gpointer callback_data)
 {
-  g_simple_async_result_complete (result);
+  _g_simple_async_result_complete_with_cancellable (result, cancellable);
+}
+
+static void
+g_daemon_file_eject_mountable_with_operation (GFile               *file,
+                                              GMountUnmountFlags   flags,
+                                              GMountOperation     *mount_operation,
+                                              GCancellable        *cancellable,
+                                              GAsyncReadyCallback  callback,
+                                              gpointer             user_data)
+{
+  guint32 dbus_flags;
+  GMountSource *mount_source;
+  const char *dbus_id, *obj_path;
+
+  mount_source = g_mount_operation_dbus_wrap (mount_operation, _g_daemon_vfs_get_async_bus ());
+
+  dbus_id = g_mount_source_get_dbus_id (mount_source);
+  obj_path = g_mount_source_get_obj_path (mount_source);
+
+  if (mount_operation)
+    g_object_ref (mount_operation);
+
+  dbus_flags = flags;
+  do_async_path_call (file,
+		      G_VFS_DBUS_MOUNT_OP_EJECT_MOUNTABLE,
+		      cancellable,
+		      callback, user_data,
+		      eject_mountable_async_cb,
+		      mount_operation, mount_operation ? g_object_unref : NULL,
+		      DBUS_TYPE_UINT32, &dbus_flags,
+                      DBUS_TYPE_STRING, &dbus_id, DBUS_TYPE_OBJECT_PATH, &obj_path,
+		      0);
+
+  g_object_unref (mount_source);
+}
+
+static gboolean
+g_daemon_file_eject_mountable_with_operation_finish (GFile               *file,
+                                                     GAsyncResult        *result,
+                                                     GError             **error)
+{
+  return TRUE;
 }
 
 static void
@@ -1342,17 +1598,7 @@ g_daemon_file_eject_mountable (GFile               *file,
 			       GAsyncReadyCallback  callback,
 			       gpointer             user_data)
 {
-  guint32 dbus_flags;
-
-  dbus_flags = flags;
-  do_async_path_call (file,
-		      G_VFS_DBUS_MOUNT_OP_EJECT_MOUNTABLE,
-		      cancellable,
-		      callback, user_data,
-		      eject_mountable_async_cb,
-		      NULL, NULL,
-		      DBUS_TYPE_UINT32, &dbus_flags,
-		      0);
+  g_daemon_file_eject_mountable_with_operation (file, flags, NULL, cancellable, callback, user_data);
 }
 
 static gboolean
@@ -1360,7 +1606,7 @@ g_daemon_file_eject_mountable_finish (GFile               *file,
 				      GAsyncResult        *result,
 				      GError             **error)
 {
-  return TRUE;
+  return g_daemon_file_eject_mountable_with_operation_finish (file, result, error);
 }
 
 static void
@@ -1370,7 +1616,83 @@ unmount_mountable_async_cb (DBusMessage *reply,
 			    GCancellable *cancellable,
 			    gpointer callback_data)
 {
-  g_simple_async_result_complete (result);
+  _g_simple_async_result_complete_with_cancellable (result, cancellable);
+}
+
+static void
+g_daemon_file_unmount_mountable_with_operation (GFile               *file,
+                                                GMountUnmountFlags   flags,
+                                                GMountOperation     *mount_operation,
+                                                GCancellable        *cancellable,
+                                                GAsyncReadyCallback  callback,
+                                                gpointer             user_data)
+{
+  guint32 dbus_flags;
+  GMountSource *mount_source;
+  const char *dbus_id, *obj_path;
+
+  mount_source = g_mount_operation_dbus_wrap (mount_operation, _g_daemon_vfs_get_async_bus ());
+
+  dbus_id = g_mount_source_get_dbus_id (mount_source);
+  obj_path = g_mount_source_get_obj_path (mount_source);
+
+  if (mount_operation)
+    g_object_ref (mount_operation);
+
+  dbus_flags = flags;
+  do_async_path_call (file,
+		      G_VFS_DBUS_MOUNT_OP_UNMOUNT_MOUNTABLE,
+		      cancellable,
+		      callback, user_data,
+		      unmount_mountable_async_cb,
+		      mount_operation, mount_operation ? g_object_unref : NULL,
+		      DBUS_TYPE_UINT32, &dbus_flags,
+                      DBUS_TYPE_STRING, &dbus_id, DBUS_TYPE_OBJECT_PATH, &obj_path,
+		      0);
+
+  g_object_unref (mount_source);
+}
+
+static gboolean
+g_daemon_file_unmount_mountable_with_operation_finish (GFile               *file,
+                                                       GAsyncResult        *result,
+                                                       GError             **error)
+{
+  return TRUE;
+}
+
+static void
+poll_mountable_async_cb (DBusMessage *reply,
+                         DBusConnection *connection,
+                         GSimpleAsyncResult *result,
+                         GCancellable *cancellable,
+                         gpointer callback_data)
+{
+  _g_simple_async_result_complete_with_cancellable (result, cancellable);
+}
+
+static void
+g_daemon_file_poll_mountable (GFile               *file,
+                              GCancellable        *cancellable,
+                              GAsyncReadyCallback  callback,
+                              gpointer             user_data)
+{
+  do_async_path_call (file,
+		      G_VFS_DBUS_MOUNT_OP_POLL_MOUNTABLE,
+		      cancellable,
+		      callback, user_data,
+		      poll_mountable_async_cb,
+		      NULL,
+		      0,
+                      0);
+}
+
+static gboolean
+g_daemon_file_poll_mountable_finish (GFile               *file,
+                                     GAsyncResult        *result,
+                                     GError             **error)
+{
+  return TRUE;
 }
 
 static void
@@ -1380,17 +1702,7 @@ g_daemon_file_unmount_mountable (GFile               *file,
 				 GAsyncReadyCallback  callback,
 				 gpointer             user_data)
 {
-  guint32 dbus_flags;
-
-  dbus_flags = flags;
-  do_async_path_call (file,
-		      G_VFS_DBUS_MOUNT_OP_UNMOUNT_MOUNTABLE,
-		      cancellable,
-		      callback, user_data,
-		      unmount_mountable_async_cb,
-		      NULL, NULL,
-		      DBUS_TYPE_UINT32, &dbus_flags,
-		      0);
+  g_daemon_file_unmount_mountable_with_operation (file, flags, NULL, cancellable, callback, user_data);
 }
 
 static gboolean
@@ -1398,14 +1710,14 @@ g_daemon_file_unmount_mountable_finish (GFile               *file,
 					GAsyncResult        *result,
 					GError             **error)
 {
-  return TRUE;
+  return g_daemon_file_unmount_mountable_with_operation_finish (file, result, error);
 }
 
-
 typedef struct {
   GFile *file;
   GMountOperation *mount_operation;
   GAsyncReadyCallback callback;
+  GCancellable *cancellable;
   gpointer user_data;
 } MountData;
 
@@ -1439,9 +1751,11 @@ mount_reply (DBusMessage *reply,
 				       g_daemon_file_mount_enclosing_volume);
     }
 
-  g_simple_async_result_complete (res);
+  _g_simple_async_result_complete_with_cancellable (res, data->cancellable);
   
   g_object_unref (data->file);
+  if (data->cancellable)
+    g_object_unref (data->cancellable);
   if (data->mount_operation)
     g_object_unref (data->mount_operation);
   g_free (data);
@@ -1481,6 +1795,8 @@ g_daemon_file_mount_enclosing_volume (GFile *location,
 
   data = g_new0 (MountData, 1);
   data->callback = callback;
+  if (data->cancellable)
+    data->cancellable = g_object_ref (data->cancellable);
   data->user_data = user_data;
   data->file = g_object_ref (location);
   if (mount_operation)
@@ -1561,7 +1877,7 @@ query_fs_info_async_cb (DBusMessage *reply,
       g_simple_async_result_set_error (result,
 				       G_IO_ERROR, G_IO_ERROR_FAILED,
 				       _("Invalid return value from query_info"));
-      g_simple_async_result_complete (result);
+      _g_simple_async_result_complete_with_cancellable (result, cancellable);
       return;
     }
 
@@ -1571,12 +1887,12 @@ query_fs_info_async_cb (DBusMessage *reply,
     {
       g_simple_async_result_set_from_error (result, error);
       g_error_free (error);
-      g_simple_async_result_complete (result);
+      _g_simple_async_result_complete_with_cancellable (result, cancellable);
       return;
     }
 
   g_simple_async_result_set_op_res_gpointer (result, info, g_object_unref);
-  g_simple_async_result_complete (result);
+  _g_simple_async_result_complete_with_cancellable (result, cancellable);
 }
 
 static void
@@ -1852,9 +2168,77 @@ g_daemon_file_query_writable_namespaces (GFile                      *file,
   list = _g_dbus_get_attribute_info_list (&iter, error);
   
   dbus_message_unref (reply);
+
+  if (list)
+    g_file_attribute_info_list_add (list,
+				    "metadata",
+				    G_FILE_ATTRIBUTE_TYPE_STRING, /* Also STRINGV, but no way express this ... */
+				    G_FILE_ATTRIBUTE_INFO_COPY_WITH_FILE |
+				    G_FILE_ATTRIBUTE_INFO_COPY_WHEN_MOVED);
   
   return list;
+}
 
+static gboolean
+set_metadata_attribute (GFile *file,
+			const char *attribute,
+			GFileAttributeType type,
+			gpointer value,
+			GCancellable *cancellable,
+			GError **error)
+{
+  GDaemonFile *daemon_file;
+  DBusMessage *message;
+  char *treename;
+  const char *metatreefile;
+  MetaTree *tree;
+  int appended;
+  gboolean res;
+
+  daemon_file = G_DAEMON_FILE (file);
+
+  treename = g_mount_spec_to_string (daemon_file->mount_spec);
+  tree = meta_tree_lookup_by_name (treename, FALSE);
+  g_free (treename);
+
+  message =
+    dbus_message_new_method_call (G_VFS_DBUS_METADATA_NAME,
+				  G_VFS_DBUS_METADATA_PATH,
+				  G_VFS_DBUS_METADATA_INTERFACE,
+				  G_VFS_DBUS_METADATA_OP_SET);
+  g_assert (message != NULL);
+  metatreefile = meta_tree_get_filename (tree);
+  _g_dbus_message_append_args (message,
+			       G_DBUS_TYPE_CSTRING, &metatreefile,
+			       G_DBUS_TYPE_CSTRING, &daemon_file->path,
+			       0);
+
+  appended = _g_daemon_vfs_append_metadata_for_set (message,
+						    tree,
+						    daemon_file->path,
+						    attribute,
+						    type,
+						    value);
+
+  res = TRUE;
+  if (appended == -1)
+    {
+      res = FALSE;
+      g_set_error (error, G_IO_ERROR,
+		   G_IO_ERROR_INVALID_ARGUMENT,
+		   _("Error setting file metadata: %s"),
+		   _("values must be string or list of strings"));
+    }
+  else if (appended > 0 &&
+      !_g_daemon_vfs_send_message_sync (message,
+					cancellable, error))
+    res = FALSE;
+
+  dbus_message_unref (message);
+
+  meta_tree_unref (tree);
+
+  return res;
 }
 
 static gboolean
@@ -1871,6 +2255,9 @@ g_daemon_file_set_attribute (GFile *file,
   dbus_uint32_t flags_dbus;
   GError *my_error;
 
+  if (g_str_has_prefix (attribute, "metadata::"))
+    return set_metadata_attribute (file, attribute, type, value_p, cancellable, error);
+
  retry:
   
   message = create_empty_message (file, G_VFS_DBUS_MOUNT_OP_SET_ATTRIBUTE, NULL, error);
@@ -1884,7 +2271,7 @@ g_daemon_file_set_attribute (GFile *file,
 				  DBUS_TYPE_UINT32,
 				  &flags_dbus);
 
-  _g_dbus_append_file_attribute (&iter, attribute, type, value_p);
+  _g_dbus_append_file_attribute (&iter, attribute, 0, type, value_p);
 
   my_error = NULL;
   reply = _g_vfs_daemon_call_sync (message,
@@ -2208,6 +2595,7 @@ g_daemon_file_monitor_file (GFile* file,
 typedef struct
 {
   GSimpleAsyncResult *result;
+  GCancellable       *cancellable;
   dbus_bool_t         can_seek;
   guint64             initial_offset;
 }
@@ -2229,7 +2617,9 @@ stream_open_cb (gint fd, StreamOpenParams *params)
   g_simple_async_result_set_op_res_gpointer (params->result, output_stream, g_object_unref);
 
 out:
-  g_simple_async_result_complete (params->result);
+  _g_simple_async_result_complete_with_cancellable (params->result, params->cancellable);
+  if (params->cancellable)
+    g_object_unref (params->cancellable);
   g_object_unref (params->result);
   g_slice_free (StreamOpenParams, params);
 }
@@ -2258,13 +2648,15 @@ append_to_async_cb (DBusMessage *reply,
     }
 
   open_params->result = g_object_ref (result);
+  if (cancellable)
+    open_params->cancellable = g_object_ref (cancellable);
   _g_dbus_connection_get_fd_async (connection, fd_id,
                                    (GetFdAsyncCallback) stream_open_cb, open_params);
   return;
 
 failure:
   g_slice_free (StreamOpenParams, open_params);
-  g_simple_async_result_complete (result);
+  _g_simple_async_result_complete_with_cancellable (result, cancellable);
 }
 
 static void
@@ -2279,6 +2671,9 @@ g_daemon_file_append_to_async (GFile                      *file,
   dbus_bool_t make_backup;
   guint32 dbus_flags;
   char *etag;
+  guint32 pid;
+
+  pid = get_pid_for_file (file);
 
   mode = 1;
   etag = "";
@@ -2294,6 +2689,7 @@ g_daemon_file_append_to_async (GFile                      *file,
                       DBUS_TYPE_STRING, &etag,
                       DBUS_TYPE_BOOLEAN, &make_backup,
                       DBUS_TYPE_UINT32, &dbus_flags,
+                      DBUS_TYPE_UINT32, &pid,
                       0);
 }
 
@@ -2342,7 +2738,7 @@ create_async_cb (DBusMessage *reply,
 
 failure:
   g_slice_free (StreamOpenParams, open_params);
-  g_simple_async_result_complete (result);
+  _g_simple_async_result_complete_with_cancellable (result, cancellable);
 }
 
 static void
@@ -2357,6 +2753,9 @@ g_daemon_file_create_async (GFile                      *file,
   dbus_bool_t make_backup;
   char *etag;
   guint32 dbus_flags;
+  guint32 pid;
+
+  pid = get_pid_for_file (file);
 
   mode = 0;
   etag = "";
@@ -2372,6 +2771,7 @@ g_daemon_file_create_async (GFile                      *file,
                       DBUS_TYPE_STRING, &etag,
                       DBUS_TYPE_BOOLEAN, &make_backup,
                       DBUS_TYPE_UINT32, &dbus_flags,
+                      DBUS_TYPE_UINT32, &pid,
                       0);
 }
 
@@ -2411,7 +2811,7 @@ enumerate_children_async_cb (DBusMessage *reply,
   g_simple_async_result_set_op_res_gpointer (result, enumerator, g_object_unref);
 
 out:
-  g_simple_async_result_complete (result);
+  _g_simple_async_result_complete_with_cancellable (result, cancellable);
 }
 
 static void
@@ -2428,7 +2828,7 @@ g_daemon_file_enumerate_children_async (GFile                      *file,
   GDaemonFileEnumerator *enumerator;
   char *uri;
 
-  enumerator = g_daemon_file_enumerator_new (file);
+  enumerator = g_daemon_file_enumerator_new (file, attributes);
   obj_path = g_daemon_file_enumerator_get_object_path (enumerator);
 
   uri = g_file_get_uri (file);
@@ -2519,7 +2919,7 @@ find_enclosing_mount_cb (GMountInfo *mount_info,
     }
 
 out:
-  g_simple_async_result_complete (data->result);
+  _g_simple_async_result_complete_with_cancellable (data->result, data->cancellable);
 
   if (my_error)
     g_error_free (my_error);
@@ -2601,7 +3001,7 @@ replace_async_cb (DBusMessage *reply,
 
 failure:
   g_slice_free (StreamOpenParams, open_params);
-  g_simple_async_result_complete (result);
+  _g_simple_async_result_complete_with_cancellable (result, cancellable);
 }
 
 static void
@@ -2617,6 +3017,9 @@ g_daemon_file_replace_async (GFile                      *file,
   dbus_bool_t dbus_make_backup = make_backup;
   guint32 dbus_flags = flags;
   guint16 mode = 2;
+  guint32 pid;
+
+  pid = get_pid_for_file (file);
   
   if (etag == NULL)
     etag = "";
@@ -2630,6 +3033,7 @@ g_daemon_file_replace_async (GFile                      *file,
                       DBUS_TYPE_STRING, &etag,
                       DBUS_TYPE_BOOLEAN, &dbus_make_backup,
                       DBUS_TYPE_UINT32, &dbus_flags,
+                      DBUS_TYPE_UINT32, &pid,
                       0);
 }
 
@@ -2676,7 +3080,7 @@ set_display_name_async_cb (DBusMessage *reply,
   g_simple_async_result_set_op_res_gpointer (result, file, g_object_unref);
 
 out:
-  g_simple_async_result_complete (result);
+  _g_simple_async_result_complete_with_cancellable (result, cancellable);
 }
 
 static void
@@ -2770,8 +3174,14 @@ g_daemon_file_file_iface_init (GFileIface *iface)
   iface->mount_mountable_finish = g_daemon_file_mount_mountable_finish;
   iface->unmount_mountable = g_daemon_file_unmount_mountable;
   iface->unmount_mountable_finish = g_daemon_file_unmount_mountable_finish;
+  iface->unmount_mountable_with_operation = g_daemon_file_unmount_mountable_with_operation;
+  iface->unmount_mountable_with_operation_finish = g_daemon_file_unmount_mountable_with_operation_finish;
   iface->eject_mountable = g_daemon_file_eject_mountable;
   iface->eject_mountable_finish = g_daemon_file_eject_mountable_finish;
+  iface->eject_mountable_with_operation = g_daemon_file_eject_mountable_with_operation;
+  iface->eject_mountable_with_operation_finish = g_daemon_file_eject_mountable_with_operation_finish;
+  iface->poll_mountable = g_daemon_file_poll_mountable;
+  iface->poll_mountable_finish = g_daemon_file_poll_mountable_finish;
   iface->query_filesystem_info = g_daemon_file_query_filesystem_info;
   iface->query_filesystem_info_async = g_daemon_file_query_filesystem_info_async;
   iface->query_filesystem_info_finish = g_daemon_file_query_filesystem_info_finish;
@@ -2787,6 +3197,10 @@ g_daemon_file_file_iface_init (GFileIface *iface)
   iface->make_symbolic_link = g_daemon_file_make_symbolic_link;
   iface->monitor_dir = g_daemon_file_monitor_dir;
   iface->monitor_file = g_daemon_file_monitor_file;
+  iface->start_mountable = g_daemon_file_start_mountable;
+  iface->start_mountable_finish = g_daemon_file_start_mountable_finish;
+  iface->stop_mountable = g_daemon_file_stop_mountable;
+  iface->stop_mountable_finish = g_daemon_file_stop_mountable_finish;
 
   /* Async operations */
 
