@@ -180,10 +180,10 @@ gvfs_archive_pop_job (GVfsArchive *archive)
     }
   else if (archive_errno (archive->archive))
     {
-      g_vfs_job_failed (archive->job,
-			G_IO_ERROR,
-			archive_errno (archive->archive),
-			archive_error_string (archive->archive));
+      g_vfs_job_failed_literal (archive->job,
+				G_IO_ERROR,
+				archive_errno (archive->archive),
+				archive_error_string (archive->archive));
     }
   else
     g_vfs_job_succeeded (archive->job);
@@ -233,6 +233,9 @@ g_vfs_backend_archive_finalize (GObject *object)
   GVfsBackendArchive *archive = G_VFS_BACKEND_ARCHIVE (object);
 
   g_assert (archive->file == NULL);
+
+  if (G_OBJECT_CLASS (g_vfs_backend_archive_parent_class)->finalize)
+    (*G_OBJECT_CLASS (g_vfs_backend_archive_parent_class)->finalize) (object);
 }
 
 static void
@@ -580,7 +583,9 @@ do_mount (GVfsBackend *backend,
 
 static void
 do_unmount (GVfsBackend *backend,
-	    GVfsJobUnmount *job)
+	    GVfsJobUnmount *job,
+            GMountUnmountFlags flags,
+            GMountSource *mount_source)
 {
   GVfsBackendArchive *ba = G_VFS_BACKEND_ARCHIVE (backend);
 
@@ -600,7 +605,26 @@ do_open_for_read (GVfsBackend *       backend,
   GVfsArchive *archive;
   struct archive_entry *entry;
   int result;
+  ArchiveFile *file;
 
+  file = archive_file_find (ba, filename);
+  if (file == NULL)
+    {
+      g_vfs_job_failed (G_VFS_JOB (job),
+		        G_IO_ERROR,
+			G_IO_ERROR_NOT_FOUND,
+			_("File doesn't exist"));
+      return;
+    }
+
+  if (g_file_info_get_file_type (file->info) == G_FILE_TYPE_DIRECTORY)
+    {
+      g_vfs_job_failed (G_VFS_JOB (job), G_IO_ERROR,
+			G_IO_ERROR_IS_DIRECTORY,
+			_("Can't open directory"));
+      return;
+    }
+  
   archive = gvfs_archive_new (ba, G_VFS_JOB (job));
 
   do
@@ -630,10 +654,10 @@ do_open_for_read (GVfsBackend *       backend,
 
   if (!gvfs_archive_in_error (archive))
     {
-      g_set_error (&archive->error,
-		   G_IO_ERROR,
-		   G_IO_ERROR_NOT_FOUND,
-		   _("File doesn't exist"));
+      g_set_error_literal (&archive->error,
+			   G_IO_ERROR,
+			   G_IO_ERROR_NOT_FOUND,
+			   _("File doesn't exist"));
     }
   gvfs_archive_finish (archive);
 }
