@@ -650,7 +650,7 @@ do_open_for_read (GVfsBackend *backend,
 
   g_vfs_ftp_task_send_and_check (&task,
                                  G_VFS_FTP_PASS_100 | G_VFS_FTP_FAIL_200,
-                                 &open_read_handlers[0],
+                                 open_read_handlers,
                                  file,
                                  NULL,
                                  "RETR %s", g_vfs_ftp_file_get_ftp_path (file));
@@ -917,7 +917,7 @@ do_enumerate (GVfsBackend *backend,
   GVfsBackendFtp *ftp = G_VFS_BACKEND_FTP (backend);
   GVfsFtpTask task = G_VFS_FTP_TASK_INIT (ftp, G_VFS_JOB (job));
   GVfsFtpFile *dir;
-  GList *list;
+  GList *list, *walk;
 
   dir = g_vfs_ftp_file_new_from_gvfs (ftp, dirname);
   list = g_vfs_ftp_dir_cache_lookup_dir (ftp->dir_cache,
@@ -934,10 +934,19 @@ do_enumerate (GVfsBackend *backend,
 
   g_vfs_ftp_task_done (&task);
 
-  g_vfs_job_enumerate_add_infos (job, list);
+  for (walk = list; walk; walk = walk->next)
+    {
+      GFileInfo *matched_info = g_file_info_new ();
+
+      /* copy into a new GFileInfo as g_vfs_job_enumerate_add_info()
+       * modifies the given GFileInfo */
+      g_file_info_copy_into (walk->data, matched_info);
+      g_vfs_job_enumerate_add_info (job, matched_info);
+      g_object_unref (matched_info);
+      g_object_unref (walk->data);
+    }
   g_vfs_job_enumerate_done (job);
 
-  g_list_foreach (list, (GFunc) g_object_unref, NULL);
   g_list_free (list);
   g_vfs_ftp_file_free (dir);
 }
@@ -1178,12 +1187,14 @@ ftp_output_stream_splice (GOutputStream *output,
               current = cancellable;
               g_clear_error (error);
               progress_callback (bytes_copied, total_size, progress_callback_data);
+              continue;
             }
           else
             {
               bytes_copied = -1;
               break;
             }
+          g_assert_not_reached();
         }
       if (n_read == 0)
         break;
@@ -1201,12 +1212,14 @@ ftp_output_stream_splice (GOutputStream *output,
                   current = cancellable;
                   g_clear_error (error);
                   progress_callback (bytes_copied, total_size, progress_callback_data);
+                  continue;
                 }
               else
                 {
                   bytes_copied = -1;
                   break;
                 }
+              g_assert_not_reached();
             }
 
           p += n_written;
@@ -1238,6 +1251,65 @@ ftp_output_stream_splice (GOutputStream *output,
   return bytes_copied;
 }
 
+static void
+do_pull_improve_error_message (GVfsFtpTask *task,
+		               GFile       *dest,
+                               gboolean     overwrite)
+{
+  GFileInfo *info;
+  GFileType file_type;
+  
+  /* There was an error opening the source, try to set a good error for it.
+   * Code taken from glib's gio/gfile.c:open_source_for_copy() function */
+
+  if (g_vfs_ftp_task_error_matches (task, G_IO_ERROR, G_IO_ERROR_IS_DIRECTORY))
+    {
+      /* The source is a directory, don't fail with WOULD_RECURSE immediately, 
+       * as that is less useful to the app. Better check for errors on the 
+       * target instead. 
+       */
+      g_vfs_ftp_task_clear_error (task);
+      
+      info = g_file_query_info (dest, G_FILE_ATTRIBUTE_STANDARD_TYPE,
+				G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
+				task->cancellable, &task->error);
+      if (info != NULL)
+	{
+	  file_type = g_file_info_get_file_type (info);
+	  g_object_unref (info);
+	  
+	  if (overwrite)
+	    {
+	      if (file_type == G_FILE_TYPE_DIRECTORY)
+		{
+		  g_set_error_literal (&task->error, G_IO_ERROR, G_IO_ERROR_WOULD_MERGE,
+                                       _("Can't copy directory over directory"));
+		  return;
+		}
+	      /* continue to would_recurse error */
+	    }
+	  else
+	    {
+	      g_set_error_literal (&task->error, G_IO_ERROR, G_IO_ERROR_EXISTS,
+                                   _("Target file already exists"));
+	      return;
+	    }
+	}
+      else
+	{
+	  /* Error getting info from target, return that error 
+           * (except for NOT_FOUND, which is no error here) 
+           */
+	  if (!g_vfs_ftp_task_error_matches (task, G_IO_ERROR, G_IO_ERROR_NOT_FOUND))
+            return;
+          g_vfs_ftp_task_clear_error (task);
+	}
+      
+      g_set_error_literal (&task->error, G_IO_ERROR, G_IO_ERROR_WOULD_RECURSE,
+                           _("Can't recursively copy directory"));
+    }
+}
+
 static void
 do_pull (GVfsBackend *         backend,
          GVfsJobPull *         job,
@@ -1248,17 +1320,34 @@ do_pull (GVfsBackend *         backend,
          GFileProgressCallback progress_callback,
          gpointer              progress_callback_data)
 {
-  static const GVfsFtpErrorFunc open_read_handlers[] = { error_550_is_directory, NULL };
+  static const GVfsFtpErrorFunc open_read_handlers[] = { error_550_is_directory,
+                                                         error_550_permission_or_not_found, 
+                                                         NULL };
   GVfsBackendFtp *ftp = G_VFS_BACKEND_FTP (backend);
   GVfsFtpTask task = G_VFS_FTP_TASK_INIT (ftp, G_VFS_JOB (job));
   GVfsFtpFile *src;
   GFile *dest;
   GInputStream *input;
   GOutputStream *output;
-  GFileInfo *info;
-  goffset total_size;
+  goffset total_size = 0;
   
+  src = g_vfs_ftp_file_new_from_gvfs (ftp, source);
   dest = g_file_new_for_path (local_path);
+
+  g_vfs_ftp_task_setup_data_connection (&task);
+  g_vfs_ftp_task_send_and_check (&task,
+                                 G_VFS_FTP_PASS_100 | G_VFS_FTP_FAIL_200,
+                                 open_read_handlers,
+                                 src,
+                                 NULL,
+                                 "RETR %s", g_vfs_ftp_file_get_ftp_path (src));
+  g_vfs_ftp_task_open_data_connection (&task);
+  if (g_vfs_ftp_task_is_in_error (&task))
+    {
+      do_pull_improve_error_message (&task, dest, flags & G_FILE_COPY_OVERWRITE);
+      goto out;
+    }
+
   if (flags & G_FILE_COPY_OVERWRITE)
     output = G_OUTPUT_STREAM (g_file_replace (dest,
                                               NULL,
@@ -1271,36 +1360,23 @@ do_pull (GVfsBackend *         backend,
                                              0,
                                              task.cancellable,
                                              &task.error));
-  g_object_unref (dest);
   if (output == NULL)
-    goto out;
-
-  src = g_vfs_ftp_file_new_from_gvfs (ftp, source);
-  if (progress_callback)
-    info = g_vfs_ftp_dir_cache_lookup_file (ftp->dir_cache, &task, src, TRUE);
-  else
-    info = NULL;
-  if (info)
     {
-      total_size = g_file_info_get_size (info);
-      g_object_unref (info);
-    }
-  else
-    total_size = 0;
-  g_vfs_ftp_task_setup_data_connection (&task);
-  g_vfs_ftp_task_send_and_check (&task,
-                                 G_VFS_FTP_PASS_100 | G_VFS_FTP_FAIL_200,
-                                 &open_read_handlers[0],
-                                 src,
-                                 NULL,
-                                 "RETR %s", g_vfs_ftp_file_get_ftp_path (src));
-  g_vfs_ftp_task_open_data_connection (&task);
-  if (g_vfs_ftp_task_is_in_error (&task))
-    {
-      g_vfs_ftp_file_free (src);
+      g_vfs_ftp_task_close_data_connection (&task);
+      g_vfs_ftp_task_receive (&task, 0, NULL);
       goto out;
     }
 
+  if (progress_callback)
+    {
+      GFileInfo *info = g_vfs_ftp_dir_cache_lookup_file (ftp->dir_cache, &task, src, TRUE);
+      if (info)
+        {
+          total_size = g_file_info_get_size (info);
+          g_object_unref (info);
+        }
+    }
+
   input = g_io_stream_get_input_stream (g_vfs_ftp_connection_get_data_stream (task.conn));
   ftp_output_stream_splice (output,
                             input,
@@ -1311,15 +1387,18 @@ do_pull (GVfsBackend *         backend,
                             &task.error);
   g_vfs_ftp_task_close_data_connection (&task);
   g_vfs_ftp_task_receive (&task, 0, NULL);
-  
+  g_object_unref (output);
+
   if (remove_source)
     {
       g_vfs_ftp_task_send (&task,
                 	   G_VFS_FTP_PASS_500,
                            "DELE %s", g_vfs_ftp_file_get_ftp_path (src));
     }
-  g_vfs_ftp_file_free (src);
+
 out:
+  g_object_unref (dest);
+  g_vfs_ftp_file_free (src);
   g_vfs_ftp_task_done (&task);
 }
 
