@@ -153,11 +153,11 @@ g_vfs_ftp_dir_cache_lookup_entry (GVfsFtpDirCache *  cache,
         	           G_IO_ERROR, G_IO_ERROR_NOT_DIRECTORY,
         		   _("The file is not a directory"));
     }
-  g_vfs_ftp_task_open_data_connection (task);
-
+  g_vfs_ftp_task_setup_data_connection (task);
   g_vfs_ftp_task_send (task,
         	       G_VFS_FTP_PASS_100 | G_VFS_FTP_FAIL_200,
                        "%s", cache->funcs->command);
+  g_vfs_ftp_task_open_data_connection (task);
   if (g_vfs_ftp_task_is_in_error (task))
     return NULL;
 
@@ -508,6 +508,7 @@ g_vfs_ftp_dir_cache_funcs_process (GInputStream *        stream,
   int type;
   GVfsFtpFile *file;
   char *line, *s;
+  gsize length;
 
   /* protect against code reorg - in current code, error never is NULL */
   g_assert (error != NULL);
@@ -516,11 +517,17 @@ g_vfs_ftp_dir_cache_funcs_process (GInputStream *        stream,
   data = g_data_input_stream_new (stream);
   /* we use LF only, because the mozilla code can handle lines ending in CR */
   g_data_input_stream_set_newline_type (data, G_DATA_STREAM_NEWLINE_TYPE_LF);
-  while ((line = g_data_input_stream_read_line (data, NULL, cancellable, error)))
+  while ((line = g_data_input_stream_read_line (data, &length, cancellable, error)))
     {
       struct list_result result = { 0, };
       GTimeVal tv = { 0, 0 };
 
+      /* strip trailing \r - ParseFTPList only removes it if the line ends in \r\n,
+       * but we stripped the \n already.
+       */
+      if (length > 0 && line[length - 1] == '\r')
+        line[--length] = '\0';
+
       g_debug ("<<%2d <<  %s\n", debug_id, line);
       type = ParseFTPList (line, &state, &result);
       if (type != 'd' && type != 'f' && type != 'l')
@@ -551,6 +558,7 @@ g_vfs_ftp_dir_cache_funcs_process (GInputStream *        stream,
       g_free (s);
       if (file == NULL)
         {
+          g_debug ("# invalid filename, skipping");
           g_free (line);
           continue;
         }
